#ifndef ARRAY_RAW_H
#define ARRAY_RAW_H

#include <cstddef>
#include <algorithm>
#include <utility>
#include <memory>

// simple dynamically allocated, fixed-size array, smart pointer
template <class T> class Array {
public:
   using value_type = T;
   using size_type = std::size_t;
   using pointer = value_type*;
   using const_pointer = const value_type*;
   using reference = value_type&;
   using const_reference = const value_type&;
   using iterator = pointer;
   using const_iterator = const_pointer;
private:
   std::unique_ptr<value_type[]> elems;
   size_type nelems{};
public:
   auto size() const noexcept { return nelems; }
   auto begin() noexcept { return elems.get(); }
   auto end() noexcept { return begin() + size(); }
   auto begin() const noexcept { return elems.get(); }
   auto end() const noexcept { return begin() + size(); }
   Array() = default;
   Array(size_type n, const_reference val)
      : elems{ new value_type[n] }, nelems{ n } {
      std::fill(begin(), end(), val);
   }
   Array(const Array &other)
      : elems{ new value_type[other.size()] }, nelems{ other.size() } {
      std::copy(other.begin(), other.end(), begin());
   }
   void swap(Array &other) noexcept {
      using std::swap;
      swap(elems, other.elems);
      swap(nelems, other.nelems);
   }
   Array& operator=(const Array &other) {
      Array{ other }.swap(*this);
      return *this;
   }
   Array(Array &&) = default;
   Array& operator=(Array &&) = default;
   ~Array() = default;
   reference operator[](size_type n) { return elems[n]; }
   const_reference operator[](size_type n) const { return elems[n]; }
   // everything else... :)
};

#endif
